# -*- coding: utf-8 -*-
#
import pygmsh
import examples

import os
import tempfile
from importlib import import_module
import subprocess


def test_generator():
    for name in examples.__all__:
        test = import_module('examples.' + name)
        yield check_output, test, name


def check_output(test, name):
    handle, filename = tempfile.mkstemp(
        prefix=name,
        suffix='.geo'
        )
    print(filename)

    with os.fdopen(handle, 'w') as h:
        h.write(test.generate())

    gmsh_out = subprocess.check_output(
        ['gmsh', '-3', filename],
        stderr=subprocess.STDOUT
        )

    return



# -*- coding: utf8 -*-
#
'''
This module provides a Python interface for the Gmsh scripting language.  It
aims at working around some of Gmsh's inconveniences (e.g., having to manually
assign an ID for every entity created) and providing access to Python's
features.
'''

# In Gmsh, the user must manually provide a unique ID for every point, curce,
# volume created. This can get messy when a lot of entities are created and it
# isn't clear which IDs are already in use. Some Gmsh commands even create new
# entities and silently reserve IDs in that way. This module tries to work
# around this by providing routines in the style of add_point(x) which
# *returns* the ID. To make variable names in Gmsh unique, keep track of how
# many points, cirlces, etc. have already been created. Variable names will
# then be p1, p2, etc. for points, c1, c2, etc. for circles and so on.
_POINT_ID = 0
_LINE_ID = 0
_LINELOOP_ID = 0
_SURFACE_ID = 0
_SURFACELOOP_ID = 0
_VOLUME_ID = 0
_CIRCLE_ID = 0
_EXTRUDE_ID = 0
_ARRAY_ID = 0
_FIELD_ID = 0

from pygmsh import __name__, __version__

def _header():
    '''Return file header.
    '''
    header = '// This code was created by %s v%s.' % (__name__, __version__)
    return header

_GMSH_CODE = [_header()]


def get_code():
    '''Returns properly formatted Gmsh code.
    '''
    return '\n'.join(_GMSH_CODE)


def Point(x, lcar):
    '''Add point.
    '''
    global _POINT_ID
    _POINT_ID += 1
    name = 'p%d' % _POINT_ID
    _GMSH_CODE.append('%s = newp;' % name)
    _GMSH_CODE.append('Point(%s) = {%g, %g, %g, %g};'
                      % (name, x[0], x[1], x[2], lcar)
                      )
    return name


def Line(p0, p1):
    '''Add line.
    '''
    global _LINE_ID
    _LINE_ID += 1
    name = 'l%d' % _LINE_ID
    _GMSH_CODE.append('%s = newl;' % name)
    _GMSH_CODE.append('Line(%s) = {%s, %s};' % (name, p0, p1))
    return name


def BSpline(control_points):
    '''Add B-spline.
    '''
    n = len(control_points)
    if n < 4:
        raise RuntimeError(('BSpline needs at least four control points (%d '
                            'given).') % n)
    global _LINE_ID
    _LINE_ID += 1
    name = 'bspline%d' % _LINE_ID
    _GMSH_CODE.append('%s = newl;' % name)
    _GMSH_CODE.append('BSpline(%s) = {%s};'
                      % (name, ', '.join(control_points))
                      )
    return name


def Circle(point_ids):
    '''Add Circle.
    '''
    global _CIRCLE_ID
    _CIRCLE_ID += 1
    name = 'c%d' % _CIRCLE_ID
    _GMSH_CODE.append('%s = newl;' % name)
    _GMSH_CODE.append('Circle(%s) = {%s, %s, %s};'
                      % (name, point_ids[0], point_ids[1], point_ids[2])
                      )
    return name


def CompoundLine(lines):
    '''Gmsh Compound Line.
    '''
    global _LINE_ID
    _LINE_ID += 1
    name = 'l%d' % _LINE_ID
    _GMSH_CODE.append('%s = newl;' % name)
    _GMSH_CODE.append('Compound Line(%s) = {%s};' % (name, ','.join(lines)))
    return name


def LineLoop(lines):
    '''Gmsh Line Loops.
    '''
    global _LINELOOP_ID
    _LINELOOP_ID += 1
    name = 'll%d' % _LINELOOP_ID
    _GMSH_CODE.append('%s = newll;' % name)
    _GMSH_CODE.append('Line Loop(%s) = {%s};' % (name, ','.join(lines)))
    return name


def PlaneSurface(line_loop):
    '''Create Gmsh Surface.
    '''
    global _SURFACE_ID
    _SURFACE_ID += 1
    sname = 'surf%d' % _SURFACE_ID
    _GMSH_CODE.append('%s = news;' % sname)
    _GMSH_CODE.append('Plane Surface(%s) = {%s};' % (sname, line_loop))
    return sname


def RuledSurface(line_loop):
    '''Create Gmsh Surface.
    '''
    global _SURFACE_ID
    _SURFACE_ID += 1
    sname = 'surf%d' % _SURFACE_ID
    _GMSH_CODE.append('%s = news;' % sname)
    _GMSH_CODE.append('Ruled Surface(%s) = {%s};' % (sname, line_loop))
    return sname


def CompoundSurface(surfaces):
    '''Gmsh Compound Surface.
    '''
    global _SURFACE_ID
    _SURFACE_ID += 1
    name = 'surf%d' % _SURFACE_ID
    _GMSH_CODE.append('%s = news;' % name)
    _GMSH_CODE.append('Compound Surface(%s) = {%s};'
                      % (name, ','.join(surfaces))
                      )
    return name


def SurfaceLoop(surfaces):
    '''Gmsh Surface Loop.
    '''
    global _SURFACELOOP_ID
    _SURFACELOOP_ID += 1
    name = 'surfloop%d' % _SURFACELOOP_ID
    _GMSH_CODE.append('%s = newsl;' % name)
    _GMSH_CODE.append('Surface Loop(%s) = {%s};' % (name, ','.join(surfaces)))
    return name


def PhysicalSurface(surface, label):
    '''Gmsh Physical Surface.
    '''
    _GMSH_CODE.append('Physical Surface("%s") = %s;' % (label, surface))
    return


def Volume(surface_loop):
    '''Gmsh Volume.
    '''
    global _VOLUME_ID
    _VOLUME_ID += 1
    name = 'vol%d' % _VOLUME_ID
    _GMSH_CODE.append('%s = newv;' % name)
    _GMSH_CODE.append('Volume(%s) = %s;' % (name, surface_loop))

    return name


def CompoundVolume(volumes):
    '''Gmsh Compound Volume.
    '''
    global _VOLUME_ID
    _VOLUME_ID += 1
    name = 'cv%d' % _VOLUME_ID
    _GMSH_CODE.append('%s = newv;' % name)
    _GMSH_CODE.append('Compound Volume(%s) = {%s};'
                      % (name, ','.join(volumes))
                      )
    return name


def PhysicalVolume(volume, label):
    '''Gmsh Physical Volume.
    '''
    _GMSH_CODE.append('Physical Volume("%s") = %s;' % (label, volume))
    return


def Extrude(entity,
            translation_axis=None,
            rotation_axis=None,
            point_on_axis=None,
            angle=None
            ):
    '''Extrusion (translation + rotation) of any entity along a given
    translation_axis, around a given rotation_axis, about a given angle. If
    one of the entities is not provided, this method will produce only
    translation or rotation.
    '''
    global _EXTRUDE_ID
    _EXTRUDE_ID += 1

    # out[] = Extrude{0,1,0}{ Line{1}; };
    name = 'ex%d' % _EXTRUDE_ID
    if translation_axis is not None and rotation_axis is not None:
        _GMSH_CODE.append(('%s[] = Extrude{{%s,%s,%s}, '
                           '{%s,%s,%s}, {%s,%s,%s}, %s}{%s;};')
                          % ((name,) +
                             tuple(translation_axis) +
                             tuple(rotation_axis) +
                             tuple(point_on_axis) +
                             (angle, entity))
                          )

    elif translation_axis is not None:
        # Only translation
        _GMSH_CODE.append('%s[] = Extrude{%s,%s,%s}{%s;};'
                          % ((name,) + tuple(translation_axis) + (entity,)))
    elif rotation_axis is not None:
        # Only rotation
        _GMSH_CODE.append('%s[] = Extrude{{%s,%s,%s}, {%s,%s,%s}, %s}{%s;};'
                          % ((name,) +
                             tuple(rotation_axis) +
                             tuple(point_on_axis) +
                             (angle, entity))
                          )

    else:
        raise RuntimeError('Specify at least translation or rotation.')
    return name


def BoundaryLayer(edges_list=None,
                  faces_list=None,
                  nodes_list=None,
                  anisomax=None,
                  hfar=None,
                  hwall_n=None,
                  hwall_t=None,
                  ratio=None,
                  thickness=None
                  ):
    '''Add boundary layer.
    '''
    # Don't use [] as default argument, cf.
    # <http://stackoverflow.com/a/113198/353337>
    if edges_list is None:
        edges_list = []
    if faces_list is None:
        faces_list = []
    if nodes_list is None:
        nodes_list = []

    global _FIELD_ID
    _FIELD_ID += 1
    name = 'field%d' % _FIELD_ID

    _GMSH_CODE.append('%s = newf;' % name)

    _GMSH_CODE.append('Field[%s] = BoundaryLayer;' % name)
    if edges_list:
        _GMSH_CODE.append('Field[%s].EdgesList = {%s};'
                          % (name, ','.join(edges_list)))
    if faces_list:
        _GMSH_CODE.append('Field[%s].FacesList = {%s};'
                          % (name, ','.join(faces_list)))
    if nodes_list:
        _GMSH_CODE.append('Field[%s].NodesList = {%s};'
                          % (name, ','.join(nodes_list)))
    if hfar:
        _GMSH_CODE.append('Field[%s].hfar= %g;' % (name, hfar))
    if hwall_t:
        _GMSH_CODE.append('Field[%s].hwall_t= %g;' % (name, hwall_t))
    if hwall_n:
        _GMSH_CODE.append('Field[%s].hwall_n= %g;' % (name, hwall_n))
    if ratio:
        _GMSH_CODE.append('Field[%s].ratio= %g;' % (name, ratio))
    if thickness:
        _GMSH_CODE.append('Field[%s].thickness= %g;' % (name, thickness))
    if anisomax:
        _GMSH_CODE.append('Field[%s].AnisoMax= %g;' % (name, anisomax))
    return name


def Array(entities):
    '''Forms a Gmsh array from a list of entities.
    '''
    global _ARRAY_ID
    _ARRAY_ID += 1
    name = 'array%d' % _ARRAY_ID
    _GMSH_CODE.append('%s[] = {%s};' % (name, ','.join(entities)))
    return name + '[]'


def Comment(string):
    '''Adds a Gmsh comment.
    '''
    _GMSH_CODE.append('// ' + string)
    return


def raw_code(list_of_strings):
    '''Add raw Gmsh code.
    '''
    for string in list_of_strings:
        _GMSH_CODE.append(string)
    return



# -*- coding: utf-8 -*-
#
__version__ = '0.2.1'
__author__ = 'Nico Schlömer'
__author_email__ = 'nico.schloemer@gmail.com'
__website__ = 'https://github.com/nschloe/pygmsh'

from pygmsh.basic import *
from pygmsh.extra import *

__all__ = [
        'basic',
        'extra'
        ]



# -*- coding: utf8 -*-
#
'''
This module contains some convenience functions for building simple geometric
objects with Gmsh.
'''

from pygmsh.basic import *
import numpy


def rotation_matrix(u, theta):
    '''Return matrix that implements the rotation around the vector :math:`u`
    by the angle :math:`\\theta`, cf.
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle.

    :param u: rotation vector
    :param theta: rotation angle
    '''
    # Cross-product matrix.
    cpm = numpy.array([[0.0,   -u[2],  u[1]],
                      [u[2],    0.0, -u[0]],
                      [-u[1],  u[0],  0.0]])
    c = numpy.cos(theta)
    s = numpy.sin(theta)
    R = numpy.eye(3) * c \
        + s * cpm \
        + (1.0 - c) * numpy.outer(u, u)
    return R


def add_rectangle(xmin, xmax, ymin, ymax, z, lcar):
    X = [[xmin, ymin, z],
         [xmax, ymin, z],
         [xmax, ymax, z],
         [xmin, ymax, z]]
    # Create points.
    p = [Point(x, lcar) for x in X]
    # Create lines
    e = [Line(p[k], p[k+1]) for k in range(len(p)-1)]
    e.append(Line(p[-1], p[0]))
    ll = LineLoop(e)
    s = PlaneSurface(ll)
    return s


def add_polygon(X, lcar):
    # Create points.
    p = [Point(x, lcar) for x in X]
    # Create lines
    e = [Line(p[k], p[k+1]) for k in range(len(p)-1)]
    e.append(Line(p[-1], p[0]))
    ll = LineLoop(e)
    s = PlaneSurface(ll)
    return s


def add_circle(radius, lcar,
               R=numpy.eye(3),
               x0=numpy.array([0.0, 0.0, 0.0]),
               compound=False,
               num_sections=3
               ):
    '''Add circle in the :math:`y`-:math:`z`-plane.
    '''
    # Define points that make the circle (midpoint and the four cardinal
    # directions).
    X = [[0.0, 0.0, 0.0]]
    if num_sections == 4:
        # For accuracy, the points are provided explicitly.
        X = [[0.0, 0.0,     0.0],
             [0.0, radius,  0.0],
             [0.0, 0.0,     radius],
             [0.0, -radius, 0.0],
             [0.0, 0.0,     -radius]]
    else:
        for k in range(num_sections):
            alpha = 2*numpy.pi * k / num_sections
            X.append([0.0, radius*numpy.cos(alpha), radius*numpy.sin(alpha)])

    # Apply the transformation.
    # TODO assert that the transformation preserves circles
    X = [numpy.dot(R, x) + x0 for x in X]
    # Add Gmsh Points.
    Comment('Points')
    p = [Point(x, lcar) for x in X]

    # Define the circle arcs.
    Comment('Circle arcs')
    c = []
    for k in range(1, len(p)-1):
        c.append(Circle([p[k], p[0], p[k+1]]))
    # Don't forget the closing arc.
    c.append(Circle([p[-1], p[0], p[1]]))
    if compound:
        c = [CompoundLine(c)]
    return c


def add_ball(x0, radius, lcar,
             with_volume=True,
             holes=None,
             label=None
             ):
    '''Creates a ball with a given radius around a given midpoint :math:`x_0`.
    '''
    if holes is None:
        holes = []

    # Add points.
    p = [Point(x0, lcar=lcar),
         Point([x0[0]+radius, x0[1],        x0[2]],        lcar=lcar),
         Point([x0[0],        x0[1]+radius, x0[2]],        lcar=lcar),
         Point([x0[0],        x0[1],        x0[2]+radius], lcar=lcar),
         Point([x0[0]-radius, x0[1],        x0[2]],        lcar=lcar),
         Point([x0[0],        x0[1]-radius, x0[2]],        lcar=lcar),
         Point([x0[0],        x0[1],        x0[2]-radius], lcar=lcar)
         ]
    # Add ball skeleton.
    c = [Circle([p[1], p[0], p[6]]),
         Circle([p[6], p[0], p[4]]),
         Circle([p[4], p[0], p[3]]),
         Circle([p[3], p[0], p[1]]),
         Circle([p[1], p[0], p[2]]),
         Circle([p[2], p[0], p[4]]),
         Circle([p[4], p[0], p[5]]),
         Circle([p[5], p[0], p[1]]),
         Circle([p[6], p[0], p[2]]),
         Circle([p[2], p[0], p[3]]),
         Circle([p[3], p[0], p[5]]),
         Circle([p[5], p[0], p[6]])
         ]
    # Add surfaces (1/8th of the ball surface).
    ll = [LineLoop([c[4],      c[9],     c[3]]),
          LineLoop([c[8],      '-'+c[4], c[0]]),
          LineLoop([c[11],     '-'+c[7], '-'+c[0]]),
          LineLoop([c[7],      '-'+c[3], c[10]]),
          LineLoop(['-'+c[9],  c[5],     c[2]]),
          LineLoop(['-'+c[10], '-'+c[2], c[6]]),
          LineLoop(['-'+c[1],  '-'+c[6], '-'+c[11]]),
          LineLoop(['-'+c[5],  '-'+c[8], c[1]])
          ]
    # Create a surface for each line loop.
    s = [RuledSurface(l) for l in ll]
    # Create the surface loop.
    surface_loop = SurfaceLoop(s)
    if holes:
        # Create an array of surface loops; the first entry is the outer
        # surface loop, the following ones are holes.
        surface_loop = Array([surface_loop] + holes)
    # Create volume.
    if with_volume:
        volume = Volume(surface_loop)
        if label:
            PhysicalVolume(volume, label)
    else:
        volume = None
    return volume, surface_loop


def add_box(x0, x1, y0, y1, z0, z1,
            lcar,
            with_volume=True,
            holes=None,
            label=None
            ):

    if holes is None:
        holes = []

    # Define corner points.
    p = [Point([x1, y1, z1], lcar=lcar),
         Point([x1, y1, z0], lcar=lcar),
         Point([x1, y0, z1], lcar=lcar),
         Point([x1, y0, z0], lcar=lcar),
         Point([x0, y1, z1], lcar=lcar),
         Point([x0, y1, z0], lcar=lcar),
         Point([x0, y0, z1], lcar=lcar),
         Point([x0, y0, z0], lcar=lcar)
         ]
    # Define edges.
    e = [Line(p[0], p[1]),
         Line(p[0], p[2]),
         Line(p[0], p[4]),
         Line(p[1], p[3]),
         Line(p[1], p[5]),
         Line(p[2], p[3]),
         Line(p[2], p[6]),
         Line(p[3], p[7]),
         Line(p[4], p[5]),
         Line(p[4], p[6]),
         Line(p[5], p[7]),
         Line(p[6], p[7])
         ]
    # Define the six line loops.
    ll = [LineLoop([e[0], e[3],  '-'+e[5],  '-'+e[1]]),
          LineLoop([e[0], e[4],  '-'+e[8],  '-'+e[2]]),
          LineLoop([e[1], e[6],  '-'+e[9],  '-'+e[2]]),
          LineLoop([e[3], e[7],  '-'+e[10], '-'+e[4]]),
          LineLoop([e[5], e[7],  '-'+e[11], '-'+e[6]]),
          LineLoop([e[8], e[10], '-'+e[11], '-'+e[9]])
          ]
    # Create a surface for each line loop.
    s = [RuledSurface(l) for l in ll]
    # Create the surface loop.
    surface_loop = SurfaceLoop(s)
    if holes:
        # Create an array of surface loops; the first entry is the outer
        # surface loop, the following ones are holes.
        surface_loop = Array([surface_loop] + holes)
    if with_volume:
        # Create volume
        vol = Volume(surface_loop)
        if label:
            PhysicalVolume(vol, label)
    else:
        vol = None
    return vol, surface_loop


def add_torus(irad, orad,
              lcar,
              R=numpy.eye(3),
              x0=numpy.array([0.0, 0.0, 0.0]),
              label=None
              ):
    '''Create Gmsh code for the torus under the coordinate transformation

    .. math::
        \hat{x} = R x + x_0.

    :param irad: inner radius of the torus
    :param orad: outer radius of the torus
    '''
    Comment(76 * '-')
    Comment('Torus')

    # Add circle
    x0t = numpy.dot(R, numpy.array([0.0, orad, 0.0]))
    c = add_circle(irad, lcar, R=R, x0=x0+x0t)

    rot_axis = [0.0, 0.0, 1.0]
    rot_axis = numpy.dot(R, rot_axis)
    point_on_rot_axis = [0.0, 0.0, 0.0]
    point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

    # Form the torus by extruding the circle three times by 2/3*pi.
    # This works around the inability of Gmsh to extrude by pi or more.  The
    # Extrude() macro returns an array; the first [0] entry in the array is
    # the entity that has been extruded at the far end. This can be used for
    # the following Extrude() step.  The second [1] entry of the array is the
    # surface that was created by the extrusion.
    previous = c
    angle = '2*Pi/3'
    all_names = []
    for i in range(3):
        Comment('Round no. %s' % (i+1))
        for k in range(len(previous)):
            # ts1[] = Extrude {{0,0,1}, {0,0,0}, 2*Pi/3}{Line{tc1};};
            # ...
            tmp_name = Extrude(
                'Line{%s}' % previous[k],
                rotation_axis=rot_axis,
                point_on_axis=point_on_rot_axis,
                angle=angle
                )
            all_names.append(tmp_name)
            previous[k] = tmp_name + '[0]'

    # Now build surface loop and volume.
    all_surfaces = [name + '[1]' for name in all_names]

    # compound_surface = CompoundSurface(all_surfaces)

    surface_loop = SurfaceLoop(all_surfaces)
    vol = Volume(surface_loop)
    if label:
        PhysicalVolume(vol, label)
    Comment(76*'-')
    return


def add_torus2(irad, orad,
               lcar,
               R=numpy.eye(3),
               x0=numpy.array([0.0, 0.0, 0.0]),
               label=None
               ):
    '''Create Gmsh code for the torus under the coordinate transformation

    .. math::
        \hat{x} = R x + x_0.

    :param irad: inner radius of the torus
    :param orad: outer radius of the torus
    '''
    Comment(76*'-')
    Comment('Torus')

    # Add circle
    x0t = numpy.dot(R, numpy.array([0.0, orad, 0.0]))
    c = add_circle(irad, lcar, R=R, x0=x0+x0t)
    ll = LineLoop(c)
    s = PlaneSurface(ll)

    rot_axis = [0.0, 0.0, 1.0]
    rot_axis = numpy.dot(R, rot_axis)
    point_on_rot_axis = [0.0, 0.0, 0.0]
    point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

    # Form the torus by extruding the circle three times by 2/3*pi.
    # This works around the inability of Gmsh to extrude by pi or more.  The
    # Extrude() macro returns an array; the first [0] entry in the array is
    # the entity that has been extruded at the far end. This can be used for
    # the following Extrude() step.  The second [1] entry of the array is the
    # surface that was created by the extrusion.
    previous = s
    all_names = []
    num_steps = 3
    for _ in range(num_steps):
        tmp_name = Extrude(
            'Surface{%s}' % previous,
            rotation_axis=rot_axis,
            point_on_axis=point_on_rot_axis,
            angle='2*Pi/%d' % num_steps
            )
        previous = tmp_name + '[0]'
        all_names.append(tmp_name)

    all_volumes = [name + '[1]' for name in all_names]
    vol = CompoundVolume(all_volumes)
    if label:
        PhysicalVolume(vol, label)
    Comment(76*'-')
    return


def add_pipe(outer_radius, inner_radius, length,
             R=numpy.eye(3),
             x0=numpy.array([0.0, 0.0, 0.0]),
             label=None,
             lcar=0.1
             ):
    '''Hollow cylinder.
    Define a rectangle, extrude it by rotation.
    '''
    Comment('Define rectangle.')
    X = numpy.array([[0.0, outer_radius, -0.5*length],
                     [0.0, outer_radius,  0.5*length],
                     [0.0, inner_radius,  0.5*length],
                     [0.0, inner_radius, -0.5*length]
                     ])
    # Apply transformation.
    X = [numpy.dot(R, x) + x0 for x in X]
    # Create points set.
    p = [Point(x, lcar) for x in X]

    # Define edges.
    e = [Line(p[0], p[1]),
         Line(p[1], p[2]),
         Line(p[2], p[3]),
         Line(p[3], p[0])
         ]

    rot_axis = [0.0, 0.0, 1.0]
    rot_axis = numpy.dot(R, rot_axis)
    point_on_rot_axis = [0.0, 0.0, 0.0]
    point_on_rot_axis = numpy.dot(R, point_on_rot_axis) + x0

    # Extrude all edges three times by 2*Pi/3.
    previous = e
    angle = '2*Pi/3'
    all_names = []
    # com = []
    Comment('Extrude in 3 steps.')
    for i in range(3):
        Comment('Step %s' % (i+1))
        for k in range(len(previous)):
            # ts1[] = Extrude {{0,0,1}, {0,0,0}, 2*Pi/3}{Line{tc1};};
            name = Extrude('Line{%s}' % previous[k],
                           rotation_axis=rot_axis,
                           point_on_axis=point_on_rot_axis,
                           angle=angle
                           )
            # if k==0:
            #     com.append(name+'[1]')
            # else:
            #     all_names.append(name+'[1]')
            all_names.append(name+'[1]')
            previous[k] = name + '[0]'
    #
    # cs = CompoundSurface(com)
    # Now just add surface loop and volume.
    all_surfaces = all_names
    # all_surfaces = all_names + [cs]
    surface_loop = SurfaceLoop(all_surfaces)
    vol = Volume(surface_loop)
    if label:
        PhysicalVolume(vol, label)
    return


def add_pipe2(outer_radius, inner_radius, length,
              R=numpy.eye(3),
              x0=numpy.array([0.0, 0.0, 0.0]),
              label=None,
              lcar=0.1
              ):
    '''Hollow cylinder.
    Define a ring, extrude it by translation.
    '''
    # Define ring which to Extrude by translation.
    c_inner = add_circle(inner_radius, lcar,
                         R=R,
                         x0=x0
                         )
    ll_inner = LineLoop(c_inner)

    c_outer = add_circle(outer_radius, lcar,
                         R=R,
                         x0=x0
                         )
    ll_outer = LineLoop(c_outer)

    surf = PlaneSurface(','.join([ll_outer, ll_inner]))

    # Now Extrude the ring surface.
    name = Extrude('Surface{%s}' % surf,
                   translation_axis=[length, 0, 0]
                   )
    vol = name + '[0]'
    if label:
        PhysicalVolume(vol, label)
    return vol
